# Heat transfer uncertainty calculations
# Variables:
    # uQsens = sensible load uncertainty
    # uQsens(w1,w2,t1,t2,p,flow,mdot,nu,df,n,SensInts,LatInts,dfHeat,dfHumid,units = 'IP',cross = 0, k = 0)
    # uQlat = latent load uncertainty
    # uQlat(w1,w2,t1,t2,p,flow,mdot,nu,df,n,SensInts,LatInts,dfHeat,dfHumid,units = 'IP',cross = 0, k = 0)
    # uQ = total load uncertainty for moist air
    # uQ(w1,w2,t1,t2,h1,h2,p,flow,mdot,nu,df,n,SensInts,LatInts,dfHeat,dfHumid,units = 'IP',cross = 0, k = 0)
    # w1, w2 are humidity ratios in and out
    # t1, t2 are the temperatures in and out in F
    # h1, h2 are the enthalpies in and out in kJ/kg
    # p is the pressure in psi
    # flow is the airflow rate in cfm
    # mdot is the mass flow rate in kg/s
    # nu is the specific volume in m3/kg
    # df is the dataframe
    # n is a label that is tacked onto the variable names. For example, '1' might be used for ZS1 data
# If you are using cross terms in calculations:
    # SensInts = steady state intervals of data containing only sensible loads
    # LatInts = steady state intervals of data containing only latent loads
    # dfHeat = data for only sensible loads
    # dfHumid = data for only latent loads
    # units: unless you convert t's to C, p to kPa, flow to m3/s, this should be IP
    # cross: if you are using the cross terms, cross = 1, otherwise 0. If you aren't using cross terms, SensInts, LatInts, dfHeat, dfHumid
    # are all []    
    # k: if you don't want to include the cross terms in the final uncertainty calculation, but you still want to calculate the cross 
    # terms, set cross = 1 and k = 0. If k = 1, then the cross terms are also included in the final uncertainty calc.
    
# If you don't have two sets of intervals, that's fine. Just enter steady state intervals in SensInts and steady state dataframe
# in dfHeat. It doesn't matter if they are specific to sensible or latent only. That's just the data I had when I first developed this.

import numpy as np
import pandas as pd
import sys
sys.path.insert(0, dirName) # dirName = location where the files listed here are located
from psychClass import Psych

class HeatTransAirUnc():
    props = Psych()
    
    def FtoC(self,t):
        '''Convert F to C - t = temperature in F'''
        return 5/9*(t - 32)

    def uRTD(self,T,unitsIn = 'IP',unitsOut = 'IP'):
        '''Calculate uncertainty of an RTD'''
        if unitsIn == 'IP':
            TC = (T-32)*5/9
        else:
            TC = T
        u = 0.15+0.002*TC # °C
        if unitsOut == 'IP':
            u = u*9/5
        return u 

    def uCard(self,unitsOut = 'IP'):
        '''Uncertainty of the RTD data acquisition card, manufacturer spec'''
        u = 0.11 # °C
        if unitsOut == 'IP':
            u = u*9/5
        return u

    def uEbtron(self,unitsOut = 'IP)'):
        '''Unceratinty of the Ebtron temperature output, manufacturer spec'''
        u = 0.09 # °C
        if unitsOut == 'IP':
            u = u*9/5
        return u

    def uTot(self,name,T,mData,unitsIn = 'IP',unitsOut = 'IP'):
        '''Total uncertainty'''
        uRTD1 = self.uRTD(T,unitsIn,'SI') # °C
        # These sensors have not been calibrated; they use only the manufacturer data
        noCalList = ['dx_c1_tsc1','dx_c1_tsc2','dx_c1_tsh1','dx_c1_tsh2','dx_c2_tsc1',
                     'dx_c2_tsc2','dx_c2_tsh1','dx_c2_tsh2','indoor_rtd','oau_cc_rtd',
                     'oau_hs_rtd','oau_in_rtd','oau_out_rtd','outdoor_rtd']
        EbtronList = ['ahu1_t_sa','ahu2_t_sa','ahu1_t_ra','ahu2_t_ra','vav1_t_in',
                      'vav2_t_in','vav3_t_in','vav4_t_in','exf1_t_in','exf2_t_in']

        if name in noCalList:
            noCal = 1
        elif name in EbtronList:
            Ebtron == 1
        else:
            noCal = 0
            Ebtron = 0

        if noCal == 1:        
            u = np.sqrt(self.uCard('SI')**2+uRTD1**2)
        elif Ebtron == 1:
            u = self.uEbtron(unitsOut)
        else:
            u = float(mData.loc[name].accuracy_scaled)*5/9

        if unitsOut == 'IP':
            u = u*9/5 # °F
            uRTD1 = uRTD1*9/5
        return u,uRTD1

    def dPsatDT(self,t,units = 'IP'): # F
        '''Derivative of the saturation pressure with respect to temperature'''
        if units == 'IP':
            tC = 5/9*(t - 32) # [C]
        else:
            tC = t        
        tK = tC + 273.15 # [K]

        C8 = -5.8002206e3
        C9 = 1.3914993
        C10 = -4.8640239e-2
        C11 = 4.1764768e-5
        C12 = -1.4452093e-8
        C13 = 6.5459673 

        pSat1 = HeatTransAirUnc.props.pSat(tC,'SI') # Pa
        a = -C8/tK**2+C10+2*C11*tK+3*C12*tK**2+C13/tK
        return a*pSat1,pSat1 # Pa/C, Pa

    def dwDT(self,RH,t,p,units = 'IP'): # %, F, psi
        '''Derivative of the humidity ratio with respect to temperature'''
        if units == 'IP':
            tC = 5/9*(t - 32) # [C]
            pSI = p*6894.76/1000 # [kPa]
        else:
            tC = t        
            pSI = p
        RH = np.array([x/100 for x in RH]) # convert from % to unitless
        dPsatDT1,pSat1 = self.dPsatDT(tC,units) # Pa/C, Pa
        pw = RH*pSat1/1000 # kPa
        return 0.622*RH*p/(p-pw)**2*dPsatDT1/1000 # kPa/C

    def dCpDT(self,RH,t,p,units = 'IP'): # kJ/kg-C * 1/C; RH [%], t [F], p [psi]
        '''Derivative of the specific heat with respect to temperature'''
        dwDT1 = self.dwDT(RH,t,p,units = 'IP') # kPa/C
        return 1.86*dwDT1 # kJ/kg-C-C

    def dMdotDT(self,t,mdot,nu,p,RH,units = 'IP'): # F, kg/s, m3/kg, psi, %
        '''Derivative of the mass flow rate with respect to temperature'''
        if units == 'IP':
            tC = 5/9*(t - 32) # [C]
            pSI = p*6894.76/1000 # [kPa]
        else:
            tC = t        
            pSI = p
        tK = tC + 273.15 # [K]
        dMdotDT1 = -mdot/nu*self.dNuDT(RH,t,p,nu,units) #
        return dMdotDT1

    def dNuDT(self,RH,t,p,nu,units = 'IP'):
        '''Derivative of the specific volume with respect to temperature'''
        dwDT1 = self.dwDT(RH,t,p,units)
        if units == 'IP':
            tC = 5/9*(t - 32) # [C]
            pSI = p*6894.76/1000 # [kPa]
        else:
            tC = t        
            pSI = p
        tK = tC + 273.15 # [K]    
        return nu/tK+0.462*tK/pSI*dwDT1

    def dwDP(self,omega,p,t,RH,units = 'IP'): # kgw/kgda, psi, F, %
        '''Derivative of the humidity ratio with respect to pressure'''
        if units == 'IP':
            tC = self.FtoC(t)
            pSI = p*6894.76/1000 # [kPa]
        else:
            tC = t
            pSI = p
        pSat1 = HeatTransAirUnc.props.pSat(tC,'SI') # Pa
        RH = np.array([x/100 for x in RH]) # convert from % to unitless
        pw = RH*pSat1/1000 # kPa
        return -omega/(pSI-pw) # kgw/kgda/kPa

    def dCpDP(self,omega,p,t,RH,units = 'IP'): # kgw/kgda,psi,F,%
        '''Derivative of the specific heat with respect to pressure'''
        return 1.86*self.dwDP(omega,p,t,RH,units) # kJ/kg-K/kPa; 1.86 has units of kJ/kg-K kgda/kgw

    def dnuDP(self,omega,p,t,RH,nu,units = 'IP'): # kgw/kgda, psi, F, %, m3/kgda
        '''Derivative of the specific volume with respect to pressure'''
        if units == 'IP':
            tC = self.FtoC(t)
            pSI = p*6894.76/1000 # [kPa]
        else:
            tC = t
            pSI = p
        tK = tC + 273.15
        return (0.462*self.dwDP(omega,p,t,RH,units)*tK-nu)/pSI # m3/kgda/kPa

    def dMdotDP(self,mdot,omega,p,t,RH,nu,units = 'IP'): # kgda/s, kgw/kgda, psi, F, %, m3/kgda
        '''Derivative of the mass flow rate with respect to pressure'''
        return -mdot/nu*self.dnuDP(omega,p,t,RH,nu,units = 'IP') # kgda/s/kPa

    def dwDRH(self,omega,p,t,RH,units = 'IP'): # kgw/kgda, psi, F, %
        '''Derivative of the humidity ratio with respect to relative humidity'''
        if units == 'IP':
            tC = self.FtoC(t)
            pSI = p*6894.76/1000 # [kPa]
        else:
            tC = t
            pSI = p
        pSat1 = HeatTransAirUnc.props.pSat(t,units) # Pa
        RH = np.array([x/100 for x in RH]) # convert from % to unitless
        pw = RH*pSat1/1000 # kPa
        return omega*pSI/(RH*(pSI-pw)) # kgw/kgda

    def dCpdRH(self,omega,p,t,RH,units = 'IP'): # kgw/kgda, psi, F, %
        '''Derivative of the specific heat with respect to relative humidity'''
        dwDRH1 = self.dwDRH(omega,p,t,RH,units) # kgw/kgda
        return 1.86*dwDRH1 # kJ/kg-K

    def dMdotDRH(self,mdot,omega,p,t,RH,nu,units = 'IP'): # kgda/s, kgw/kgda, psi, F, %, m3/kgda
        '''Derivative of the mass flow rate with respect to relative humidity'''
        dNuDRH1 = self.dNuDRH(omega,p,t,RH,units) # m3/kgda
        return -mdot/nu*dNuDRH1 # kgda/s

    def dNuDRH(self,omega,p,t,RH,units = 'IP'): # kgw/kgda, psi, F, %
        '''Derivative of the specific volume with respect to relative humidity'''
        if units == 'IP':
            tC = self.FtoC(t)
            pSI = p*6894.76/1000 # [kPa]
        else:
            tC = t
            pSI = p
        tK = tC + 273.15 # [K]     
        dwDRH1 = self.dwDRH(omega,p,t,RH,units) # kgw/kgda
        return 0.462*tK/pSI*dwDRH1 # m3/kgda

    def dMdotAV(self,nu):
        '''Derivative of the mass flow rate with respect to airflow rate (cfm)'''
        return 0.0004719/nu

    def dHdT(self,omega,p,t,RH,units = 'IP'):
        '''Derivative of enthalpy with respect to temperature'''
        if units == 'IP':
            tC = self.FtoC(t)
        else:
            tC = t
        dwDT1 = self.dwDT(RH,t,p,units)
        return 1.006+1.86*omega+dwDT1*(2501+1.86*tC)

    def dHdP(self,omega,p,t,RH,units = 'IP'):
        '''Derivative of enthalpy with respect to pressure'''
        if units == 'IP':
            tC = self.FtoC(t)
        else:
            tC = t
        dwDP1 = self.dwDP(omega,p,t,RH,units)
        return dwDP1*(2501+1.86*tC)

    def dHdRH(self,omega,p,t,RH,units = 'IP'):
        '''Derivative of enthalpy with respect to relative humidity'''
        if units == 'IP':
            tC = self.FtoC(t)
        else:
            tC = t
        dwDRH1 = self.dwDRH(omega,p,t,RH,units)
        return dwDRH1*(2501+1.86*tC)


    def covEst(self,SensInts,LatInts,dfHeat,dfHumid):
        '''Calculate covariance for steady state data'''
        NHeatInts = np.shape(SensInts)[0]
        NHumidInts = np.shape(LatInts)[0]
        uTP = []
        uTRH = []
        uTAV = []
        upRH = []
        upAV = []
        uRHAV = []
        if NHeatInts > 0:
            for a,b in SensInts:
                w1 = dfHeat.w_zs1_in.iloc[a:b]
                w2 = dfHeat.w_zs2_in.iloc[a:b]
                t1 = dfHeat.vav4_out_rtd.iloc[a:b]
                t2 = dfHeat.zs4_out_rtd.iloc[a:b]
                p = dfHeat.indoor_p.iloc[a:b]
                units = 'IP'
                flow = dfHeat.vav4_f.iloc[a:b]
                mdot = dfHeat.m_dot_vav4.iloc[a:b]

                cp1 = HeatTransAirUnc.props.cp(w1,t1,units)
                cp2 = HeatTransAirUnc.props.cp(w2,t2,units)
                rh1 = HeatTransAirUnc.props.relHum(t1,w1,p,units)
                rh1 = np.array([100*x for x in rh1])
                rh2 = HeatTransAirUnc.props.relHum(t2,w2,p,units)
                rh2 = np.array([100*x for x in rh2])
                N = b-a
                if N > 0:
                    uTP.append(1/N*np.sum( (self.FtoC(t2) - np.mean(self.FtoC(t2)))*(p*6.89 - np.mean(p*6.89)) )  ) 
                    uTRH.append(1/N*np.sum( (self.FtoC(t2) - np.mean(self.FtoC(t2)))*(rh2*0.01 - np.mean(rh2*0.01)) ))
                    uTAV.append(1/N*np.sum( (self.FtoC(t2) - np.mean(self.FtoC(t2)))*(flow - np.mean(flow)) ))
                    upRH.append(1/N*np.sum( (p*6.89 - np.mean(p*6.89))*(rh2*0.01 - np.mean(rh2*0.01)) ))
                    upAV.append(1/N*np.sum( (p*6.89 - np.mean(p*6.89))*(flow - np.mean(flow)) ))
                    uRHAV.append(1/N*np.sum( (rh2*0.01 - np.mean(rh2*0.01))*(flow - np.mean(flow)) ))

        if NHumidInts > 0:
            for a,b in LatInts:
                w1 = dfHumid.w_zs1_in.iloc[a:b]
                w2 = dfHumid.w_zs2_in.iloc[a:b]
                t1 = dfHumid.vav4_out_rtd.iloc[a:b]
                t2 = dfHumid.zs4_out_rtd.iloc[a:b]
                p = dfHumid.indoor_p.iloc[a:b]
                units = 'IP'
                flow = dfHumid.vav4_f.iloc[a:b]
                mdot = dfHumid.m_dot_vav4.iloc[a:b]

                cp1 = HeatTransAirUnc.props.cp(w1,t1,units)
                cp2 = HeatTransAirUnc.props.cp(w2,t2,units)
                rh1 = HeatTransAirUnc.props.relHum(t1,w1,p,units)
                rh1 = np.array([100*x for x in rh1])
                rh2 = HeatTransAirUnc.props.relHum(t2,w2,p,units)
                rh2 = np.array([100*x for x in rh2])
                N = b-a
                if N > 0:
                    uTP.append(1/N*np.sum( (self.FtoC(t2) - np.mean(self.FtoC(t2)))*(p*6.89 - np.mean(p*6.89)) )  ) 
                    uTRH.append(1/N*np.sum( (self.FtoC(t2) - np.mean(self.FtoC(t2)))*(rh2*0.01 - np.mean(rh2*0.01)) ))
                    uTAV.append(1/N*np.sum( (self.FtoC(t2) - np.mean(self.FtoC(t2)))*(flow - np.mean(flow)) ))
                    upRH.append(1/N*np.sum( (p*6.89 - np.mean(p*6.89))*(rh2*0.01 - np.mean(rh2*0.01)) ))
                    upAV.append(1/N*np.sum( (p*6.89 - np.mean(p*6.89))*(flow - np.mean(flow)) ))
                    uRHAV.append(1/N*np.sum( (rh2*0.01 - np.mean(rh2*0.01))*(flow - np.mean(flow)) ))

        return np.median(uTP),np.median(uTRH),np.median(uTAV),np.median(upRH),np.median(upAV),np.median(uRHAV)

    def uQsens(self,w1n,w2n,t1n,t2n,pn,flown,mdotn,nun,df,mData,n,SensInts,LatInts,dfHeat,dfHumid,units = 'IP',cross = 0, k = 0):
        '''Calculate the uncertainty of sensible heat transfer in air''''
        w1 = df[w1n]
        w2 = df[w2n]
        t1 = df[t1n]
        t2 = df[t2n]
        p = df[pn]
        flow = 0
        for f in flown:
            flow = flow + df[f]
        mdot = df[mdotn]
        nu = df[nun]

        cp1 = HeatTransAirUnc.props.cp(w1,t1,units)
        cp2 = HeatTransAirUnc.props.cp(w2,t2,units)
        rh1 = HeatTransAirUnc.props.relHum(t1,w1,p,units)
        rh1 = np.array([100*x for x in rh1])
        rh2 = HeatTransAirUnc.props.relHum(t2,w2,p,units)
        rh2 = np.array([100*x for x in rh2])

        uT,_ = self.uTot(t2n,t2,mData,unitsIn = 'IP',unitsOut = 'SI')
        up = 0.05*p*6.89 # kPa, p is in psi
        uRH = 0.02 # -
        uAV = 0.02*flow # cfm

        dCpDT1 = self.dCpDT(rh1,t1,p,units)
        dCpDT2 = self.dCpDT(rh2,t2,p,units)
        term1a = ((self.FtoC(t2)*dCpDT2+cp2)-(self.FtoC(t1)*dCpDT1+cp1))*mdot
        dMdotDT1 = self.dMdotDT(t2,mdot,nu,p,rh2,units)
        term2a = dMdotDT1*(cp2*self.FtoC(t2)-cp1*self.FtoC(t1))
        dQsens_dT = term1a + term2a

        dCpDP1 = self.dCpDP(w1,p,t1,rh1,units)
        dCpDP2 = self.dCpDP(w2,p,t2,rh2,units)
        term1b = ((self.FtoC(t2)*dCpDP2)-(self.FtoC(t1)*dCpDP1))*mdot
        dMdotDP1 = self.dMdotDP(mdot,w2,p,t2,rh2,nu,units) 
        term2b = dMdotDP1*(cp2*self.FtoC(t2)-cp1*self.FtoC(t1))
        dQsens_dP = term1b+term2b

        dCpDRH1 = self.dCpdRH(w1,p,t1,rh1,units)
        dCpDRH2 = self.dCpdRH(w2,p,t2,rh2,units)
        term1c = ((self.FtoC(t2)*dCpDRH2)-(self.FtoC(t1)*dCpDRH1))*mdot
        dMdotDRH1 = self.dMdotDRH(mdot,w2,p,t2,rh2,nu,units)
        term2c = dMdotDRH1*(cp2*self.FtoC(t2)-cp1*self.FtoC(t1))
        dQsens_dRH = term1c+term2c

        dQsens_dAV = self.dMdotAV(nu)*(cp2*self.FtoC(t2)-cp1*self.FtoC(t1))

        # Terms for the first part of the first order Taylor series expansion propagation of uncertainty
        Tterm = (dQsens_dT*uT)**2
        Pterm = (dQsens_dP*up)**2
        RHterm = (dQsens_dRH*uRH)**2
        AVterm = (dQsens_dAV*uAV)**2
        uterm1 = Tterm + Pterm + RHterm + AVterm

        # Terms for the second part of the Taylor series expansion; accounts for interdependence between variables
        if cross == 1:
            uTP,uTRH,uTAV,upRH,upAV,uRHAV = self.covEst(SensInts,LatInts,dfHeat,dfHumid)
            TPterm = dQsens_dT*dQsens_dP*uTP
            TRHterm = dQsens_dT*dQsens_dRH*uTRH
            TAVterm = dQsens_dT*dQsens_dAV*uTAV
            pRHterm = dQsens_dP*dQsens_dRH*upRH
            pAVterm = dQsens_dP*dQsens_dAV*upAV
            RHAVterm = dQsens_dRH*dQsens_dAV*uRHAV
            uQSens2 = 2*(TPterm+TRHterm+TAVterm+pRHterm+pAVterm+RHAVterm) 
        else:
            TPterm = 0
            TRHterm = 0
            TAVterm = 0
            pRHterm = 0
            pAVterm = 0
            RHAVterm = 0
            uQSens2 = 0 
            
        # Final calculation of the entire uncertainty term, squared
        uQSensSq = (Tterm+Pterm+RHterm+AVterm+uQSens2*k)        
        df['u_'+n+'_sens'] = uQSensSq
        df['Tterm'+n] = Tterm
        df['Pterm'+n] = Pterm
        df['RHterm'+n] = RHterm
        df['AVterm'+n] = AVterm
        df['uQSens2'+n] = uQSens2
        df['TPterm'+n] = TPterm
        df['TRHterm'+n] = TRHterm
        df['TAVterm'+n] = TAVterm
        df['pRHterm'+n] = pRHterm
        df['pAVterm'+n] = pAVterm
        df['RHAVterm'+n] = RHAVterm
        df['dQsens_dT'+n] = dQsens_dT
        df['dQsens_dP'+n] = dQsens_dP
        df['dQsens_dRH'+n] = dQsens_dRH
        df['dQsens_dAV'+n] = dQsens_dAV
        return df # kW2 in all cases

    def uQlat(self,w1n,w2n,t1n,t2n,pn,flown,mdotn,nun,df,mData,n,SensInts,LatInts,dfHeat,dfHumid,units = 'IP',cross = 0, k = 0):
        '''Calculate the uncertainty of latent heat transfer in air''''
        w1 = df[w1n]
        w2 = df[w2n]
        t1 = df[t1n]
        t2 = df[t2n]
        p = df[pn]
        flow = 0
        for f in flown:
            flow = flow + df[f]
        mdot = df[mdotn]
        nu = df[nun]

        cp1 = HeatTransAirUnc.props.cp(w1,t1,units)
        cp2 = HeatTransAirUnc.props.cp(w2,t2,units)
        rh1 = HeatTransAirUnc.props.relHum(t1,w1,p,units)
        rh1 = np.array([100*x for x in rh1])
        rh2 = HeatTransAirUnc.props.relHum(t2,w2,p,units)
        rh2 = np.array([100*x for x in rh2])    

        uT,_ = self.uTot(t2n,t2,mData,unitsIn = 'IP',unitsOut = 'SI')
        up = 0.05*p*6.89 # kPa, p is in psi
        uRH = 0.02 # -
        uAV = 0.02*flow # cfm

        dwDT1 = self.dwDT(rh1,t1,p,units) 
        dwDT2 = self.dwDT(rh2,t2,p,units)
        term1a = (dwDT2-dwDT1)*mdot
        dMdotDT1 = self.dMdotDT(t2,mdot,nu,p,rh2,units)
        term2a = dMdotDT1*(w2-w1)
        dQlat_dT = 2470*(term1a + term2a)

        dwDP1 = self.dwDP(w1,p,t1,rh1,units) 
        dwDP2 = self.dwDP(w2,p,t2,rh2,units)
        term1b = (dwDP2-dwDP1)*mdot
        dMdotDP1 = self.dMdotDP(mdot,w2,p,t2,rh2,nu,units) 
        term2b = dMdotDP1*(w2-w1)
        dQlat_dP = 2470*(term1b+term2b)

        dwDRH1 = self.dwDRH(w1,p,t1,rh1,units) 
        dwDRH2 = self.dwDRH(w2,p,t2,rh2,units)
        term1c = (dwDRH2-dwDRH1)*mdot
        dMdotDRH1 = self.dMdotDRH(mdot,w2,p,t2,rh2,nu,units)
        term2c = dMdotDRH1*(w2-w1)
        dQlat_dRH = 2470*(term1c+term2c)

        dQlat_dAV = 2470*self.dMdotAV(nu)*(w2-w1)

        # Terms for the first part of the first order Taylor series expansion propagation of uncertainty
        Tterm = (dQlat_dT*uT)**2
        Pterm = (dQlat_dP*up)**2
        RHterm = (dQlat_dRH*uRH)**2
        AVterm = (dQlat_dAV*uAV)**2
        uterm1 = Tterm + Pterm + RHterm + AVterm

        # Terms for the second part of the Taylor series expansion; accounts for interdependence between variables
        if cross == 1:
            uTP,uTRH,uTAV,upRH,upAV,uRHAV = self.covEst(SensInts,LatInts,dfHeat,dfHumid)
            TPterm = dQlat_dT*dQlat_dP*uTP
            TRHterm = dQlat_dT*dQlat_dRH*uTRH
            TAVterm = dQlat_dT*dQlat_dAV*uTAV
            pRHterm = dQlat_dP*dQlat_dRH*upRH
            pAVterm = dQlat_dP*dQlat_dAV*upAV
            RHAVterm = dQlat_dRH*dQlat_dAV*uRHAV
            uQLat2 = 2*(TPterm+TRHterm+TAVterm+pRHterm+pAVterm+RHAVterm)    
        else:
            TPterm = 0
            TRHterm = 0
            TAVterm = 0
            pRHterm = 0
            pAVterm = 0
            RHAVterm = 0
            uQLat2 = 0  
            
        # Final calculation of the entire uncertainty term, squared
        uQLatSq = (Tterm+Pterm+RHterm+AVterm+uQLat2*k)
        df['u_'+n+'_lat'] = uQLatSq
        df['TtermLat'+n] = Tterm
        df['PtermLat'+n] = Pterm
        df['RHtermLat'+n] = RHterm
        df['AVtermLat'+n] = AVterm
        df['uQLat2'+n] = uQLat2
        df['TPtermLat'+n] = TPterm
        df['TRHtermLat'+n] = TRHterm
        df['TAVtermLat'+n] = TAVterm
        df['pRHtermLat'+n] = pRHterm
        df['pAVtermLat'+n] = pAVterm
        df['RHAVtermLat'+n] = RHAVterm
        df['dQlat_dT'+n] = dQlat_dT
        df['dQlat_dP'+n] = dQlat_dP
        df['dQlat_dRH'+n] = dQlat_dRH
        df['dQlat_dAV'+n] = dQlat_dAV
        df['lat_dwDRH1'+n] = dwDRH1
        df['lat_dwDRH2'+n] = dwDRH2
        df['lat_dMdotDRH1'+n] = dMdotDRH1
        df['lat_term1c'+n] = term1c
        df['lat_term2c'+n] = term2c
        return df # kW2 in all cases

    def uQ(self,w1n,w2n,t1n,t2n,h1n,h2n,pn,flown,mdotn,nun,df,mData,n,SensInts,LatInts,dfHeat,dfHumid,units = 'IP',cross = 0, k = 0):
        '''Calculate the uncertainty of overall heat transfer in air''''
        w1 = df[w1n]
        w2 = df[w2n]
        t1 = df[t1n]
        t2 = df[t2n]
        p = df[pn]
        if np.shape(flown)[0] > 1:
            flow = 0
            for f in flown:
                flow = flow + df[f]
        else:
            flow = df[flown]
            
        mdot = df[mdotn]
        nu = df[nun]
        h1 = df[h1n]
        h2 = df[h2n]    

        cp1 = HeatTransAirUnc.props.cp(w1,t1,units)
        cp2 = HeatTransAirUnc.props.cp(w2,t2,units)
        rh1 = HeatTransAirUnc.props.relHum(t1,w1,p,units)
        rh1 = np.array([100*x for x in rh1])
        rh2 = HeatTransAirUnc.props.relHum(t2,w2,p,units)
        rh2 = np.array([100*x for x in rh2])

        #uT = 0.12 # C = 0.21 f
        uT,_ = self.uTot(t2n,t2,mData,unitsIn = 'IP',unitsOut = 'SI')
        up = 0.05*p*6.89 # kPa, p is in psi
        uRH = 0.02 # -
        uAV = 0.02*flow # cfm

        dhDT1 = self.dHdT(w1,p,t1,rh1,units)
        dhDT2 = self.dHdT(w1,p,t2,rh2,units)
        term1a = (dhDT2-dhDT1)*mdot
        dMdotDT1 = self.dMdotDT(t2,mdot,nu,p,rh2,units)
        term2a = dMdotDT1*(h2-h1)
        dQ_dT = (term1a + term2a)

        dhDP1 = self.dHdP(w1,p,t1,rh1,units) 
        dhDP2 = self.dHdP(w2,p,t2,rh2,units)
        term1b = (dhDP2-dhDP1)*mdot
        dMdotDP1 = self.dMdotDP(mdot,w2,p,t2,rh2,nu,units) 
        term2b = dMdotDP1*(h2-h1)
        dQ_dP = (term1b+term2b)

        dhDRH1 = self.dHdRH(w1,p,t1,rh1,units) 
        dhDRH2 = self.dHdRH(w2,p,t2,rh2,units)
        term1c = (dhDRH2-dhDRH1)*mdot
        dMdotDRH1 = self.dMdotDRH(mdot,w2,p,t2,rh2,nu,units)
        term2c = dMdotDRH1*(h2-h1)
        dQ_dRH = (term1c+term2c)

        dQ_dAV = self.dMdotAV(nu)*(h2-h1)

        # Terms for the first part of the first order Taylor series expansion propagation of uncertainty
        Tterm = (dQ_dT*uT)**2
        Pterm = (dQ_dP*up)**2
        RHterm = (dQ_dRH*uRH)**2
        AVterm = (dQ_dAV*uAV)**2
        uterm1 = Tterm + Pterm + RHterm + AVterm

        # Terms for the second part of the Taylor series expansion; accounts for interdependence between variables
        if cross == 1:
            uTP,uTRH,uTAV,upRH,upAV,uRHAV = self.covEst(SensInts,LatInts,dfHeat,dfHumid)
            TPterm = dQ_dT*dQ_dP*uTP
            TRHterm = dQ_dT*dQ_dRH*uTRH
            TAVterm = dQ_dT*dQ_dAV*uTAV
            pRHterm = dQ_dP*dQ_dRH*upRH
            pAVterm = dQ_dP*dQ_dAV*upAV
            RHAVterm = dQ_dRH*dQ_dAV*uRHAV
            uQ2 = 2*(TPterm+TRHterm+TAVterm+pRHterm+pAVterm+RHAVterm)    
        else:
            TPterm = 0
            TRHterm = 0
            TAVterm = 0
            pRHterm = 0
            pAVterm = 0
            RHAVterm = 0
            uQ2 = 0
            
        # Final calculation of the entire uncertainty term, squared
        uQSq = (Tterm+Pterm+RHterm+AVterm+uQ2*k)
        df['u_'+n] = uQSq
        df['TtermT'+n] = Tterm
        df['PtermT'+n] = Pterm
        df['RHtermT'+n] = RHterm
        df['AVtermT'+n] = AVterm
        df['uQ2'+n] = uQ2
        df['TPtermT'+n] = TPterm
        df['TRHtermT'+n] = TRHterm
        df['TAVtermT'+n] = TAVterm
        df['pRHtermT'+n] = pRHterm
        df['pAVtermT'+n] = pAVterm
        df['RHAVtermT'+n] = RHAVterm

        df['dQ_dT'+n] = dQ_dT
        df['dQ_dP'+n] = dQ_dP
        df['dQ_dRH'+n] = dQ_dRH
        df['dQ_dAV'+n] = dQ_dAV
        return df # kW2 in all cases